# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class LogDownloadResult(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'log_chunks': 'list[LogFileChunk]',
        'next_page_token': 'str',
        'previous_page_token': 'str',
        'bearer_token': 'str'
    }

    attribute_map = {
        'log_chunks': 'log_chunks',
        'next_page_token': 'next_page_token',
        'previous_page_token': 'previous_page_token',
        'bearer_token': 'bearer_token'
    }

    def __init__(self, log_chunks=None, next_page_token=None, previous_page_token=None, bearer_token=None, local_vars_configuration=None):  # noqa: E501
        """LogDownloadResult - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._log_chunks = None
        self._next_page_token = None
        self._previous_page_token = None
        self._bearer_token = None
        self.discriminator = None

        self.log_chunks = log_chunks
        if next_page_token is not None:
            self.next_page_token = next_page_token
        if previous_page_token is not None:
            self.previous_page_token = previous_page_token
        if bearer_token is not None:
            self.bearer_token = bearer_token

    @property
    def log_chunks(self):
        """Gets the log_chunks of this LogDownloadResult.  # noqa: E501

        A list of log file chunks.  # noqa: E501

        :return: The log_chunks of this LogDownloadResult.  # noqa: E501
        :rtype: list[LogFileChunk]
        """
        return self._log_chunks

    @log_chunks.setter
    def log_chunks(self, log_chunks):
        """Sets the log_chunks of this LogDownloadResult.

        A list of log file chunks.  # noqa: E501

        :param log_chunks: The log_chunks of this LogDownloadResult.  # noqa: E501
        :type: list[LogFileChunk]
        """
        if self.local_vars_configuration.client_side_validation and log_chunks is None:  # noqa: E501
            raise ValueError("Invalid value for `log_chunks`, must not be `None`")  # noqa: E501

        self._log_chunks = log_chunks

    @property
    def next_page_token(self):
        """Gets the next_page_token of this LogDownloadResult.  # noqa: E501

        Page token to fetch newer logs.  # noqa: E501

        :return: The next_page_token of this LogDownloadResult.  # noqa: E501
        :rtype: str
        """
        return self._next_page_token

    @next_page_token.setter
    def next_page_token(self, next_page_token):
        """Sets the next_page_token of this LogDownloadResult.

        Page token to fetch newer logs.  # noqa: E501

        :param next_page_token: The next_page_token of this LogDownloadResult.  # noqa: E501
        :type: str
        """

        self._next_page_token = next_page_token

    @property
    def previous_page_token(self):
        """Gets the previous_page_token of this LogDownloadResult.  # noqa: E501

        Page token to fetch older logs.  # noqa: E501

        :return: The previous_page_token of this LogDownloadResult.  # noqa: E501
        :rtype: str
        """
        return self._previous_page_token

    @previous_page_token.setter
    def previous_page_token(self, previous_page_token):
        """Sets the previous_page_token of this LogDownloadResult.

        Page token to fetch older logs.  # noqa: E501

        :param previous_page_token: The previous_page_token of this LogDownloadResult.  # noqa: E501
        :type: str
        """

        self._previous_page_token = previous_page_token

    @property
    def bearer_token(self):
        """Gets the bearer_token of this LogDownloadResult.  # noqa: E501

        Token to be used as a Bearer token passed in via Authorization header to access the log chunks.  # noqa: E501

        :return: The bearer_token of this LogDownloadResult.  # noqa: E501
        :rtype: str
        """
        return self._bearer_token

    @bearer_token.setter
    def bearer_token(self, bearer_token):
        """Sets the bearer_token of this LogDownloadResult.

        Token to be used as a Bearer token passed in via Authorization header to access the log chunks.  # noqa: E501

        :param bearer_token: The bearer_token of this LogDownloadResult.  # noqa: E501
        :type: str
        """

        self._bearer_token = bearer_token

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogDownloadResult):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogDownloadResult):
            return True

        return self.to_dict() != other.to_dict()
